@@ -6,8 +6,8 @@
  *  FSTYPE - filesystem type
  *  FSSIZE - filesystem size (if known)
  *
- * We currently detect (in order):
- *  gzip, cramfs, romfs, xfs, luks, minix, ext3, ext2, reiserfs, jfs, swap
+ * We currently detect the filesystems listed below in the struct
+ * "imagetype images" (in the order they are listed).
  *
  * MINIX, ext3 and Reiserfs bits are currently untested.
  */
@@ -30,6 +30,7 @@
 #include "ext3_fs.h"
 #include "xfs_sb.h"
 #include "luks_fs.h"
+#include "lvm2_sb.h"
 
 /*
  * Slightly cleaned up version of jfs_superblock to
@@ -52,16 +53,20 @@
 /* Swap needs the definition of block size */
 #include "swap_fs.h"
 
-static int gzip_image(const unsigned char *buf, unsigned long long *bytes)
+static int gzip_image(const void *buf, unsigned long long *bytes)
 {
-	if (buf[0] == 037 && (buf[1] == 0213 || buf[1] == 0236)) {
-		*bytes = 0;	/* Can only be determined by processing the whole file */
+	const unsigned char *p = buf;
+
+	if (p[0] == 037 && (p[1] == 0213 || p[1] == 0236)) {
+		/* The length of a gzip stream can only be determined
+		   by processing the whole stream */
+		*bytes = 0ULL;
 		return 1;
 	}
 	return 0;
 }
 
-static int cramfs_image(const unsigned char *buf, unsigned long long *bytes)
+static int cramfs_image(const void *buf, unsigned long long *bytes)
 {
 	const struct cramfs_super *sb =
 		(const struct cramfs_super *)buf;
@@ -76,7 +81,7 @@ static int cramfs_image(const unsigned char *buf, unsigned long long *bytes)
 	return 0;
 }
 
-static int romfs_image(const unsigned char *buf, unsigned long long *bytes)
+static int romfs_image(const void *buf, unsigned long long *bytes)
 {
         const struct romfs_super_block *sb =
 		(const struct romfs_super_block *)buf;
@@ -88,21 +93,21 @@ static int romfs_image(const unsigned char *buf, unsigned long long *bytes)
 	return 0;
 }
 
-static int minix_image(const unsigned char *buf, unsigned long long *bytes)
+static int minix_image(const void *buf, unsigned long long *bytes)
 {
 	const struct minix_super_block *sb =
 		(const struct minix_super_block *)buf;
 
 	if (sb->s_magic == MINIX_SUPER_MAGIC ||
 	    sb->s_magic == MINIX_SUPER_MAGIC2) {
-		*bytes = (unsigned long long)sb->s_nzones 
+		*bytes = (unsigned long long)sb->s_nzones
 			<< (sb->s_log_zone_size+10);
 		return 1;
 	}
 	return 0;
 }
 
-static int ext3_image(const unsigned char *buf, unsigned long long *bytes)
+static int ext3_image(const void *buf, unsigned long long *bytes)
 {
 	const struct ext3_super_block *sb =
 		(const struct ext3_super_block *)buf;
@@ -116,7 +121,7 @@ static int ext3_image(const unsigned char *buf, unsigned long long *bytes)
 	return 0;
 }
 
-static int ext2_image(const unsigned char *buf, unsigned long long *bytes)
+static int ext2_image(const void *buf, unsigned long long *bytes)
 {
 	const struct ext2_super_block *sb =
 		(const struct ext2_super_block *)buf;
@@ -129,7 +134,7 @@ static int ext2_image(const unsigned char *buf, unsigned long long *bytes)
 	return 0;
 }
 
-static int reiserfs_image(const unsigned char *buf, unsigned long long *bytes)
+static int reiserfs_image(const void *buf, unsigned long long *bytes)
 {
 	const struct reiserfs_super_block *sb =
 		(const struct reiserfs_super_block *)buf;
@@ -147,7 +152,7 @@ static int reiserfs_image(const unsigned char *buf, unsigned long long *bytes)
 	return 0;
 }
 
-static int xfs_image(const unsigned char *buf, unsigned long long *bytes)
+static int xfs_image(const void *buf, unsigned long long *bytes)
 {
 	const struct xfs_sb *sb =
 		(const struct xfs_sb *)buf;
@@ -160,7 +165,7 @@ static int xfs_image(const unsigned char *buf, unsigned long long *bytes)
 	return 0;
 }
 
-static int jfs_image(const unsigned char *buf, unsigned long long *bytes)
+static int jfs_image(const void *buf, unsigned long long *bytes)
 {
 	const struct jfs_superblock *sb =
 		(const struct jfs_superblock *)buf;
@@ -172,7 +177,7 @@ static int jfs_image(const unsigned char *buf, unsigned long long *bytes)
 	return 0;
 }
 
-static int luks_image(const unsigned char *buf, unsigned long long *blocks)
+static int luks_image(const void *buf, unsigned long long *blocks)
 {
 	const struct luks_partition_header *lph =
 		(const struct luks_partition_header *)buf;
@@ -185,7 +190,7 @@ static int luks_image(const unsigned char *buf, unsigned long long *blocks)
 	return 0;
 }
 
-static int swap_image(const unsigned char *buf, unsigned long long *blocks)
+static int swap_image(const void *buf, unsigned long long *blocks)
 {
 	const struct swap_super_block *ssb =
 		(const struct swap_super_block *)buf;
@@ -198,24 +203,56 @@ static int swap_image(const unsigned char *buf, unsigned long long *blocks)
 	return 0;
 }
 
+static int lvm2_image(const void *buf, unsigned long long *blocks)
+{
+	const struct lvm2_super_block *lsb;
+	int i;
+
+	/* We must check every 512 byte sector */
+	for (i = 0; i < BLOCK_SIZE; i += 0x200) {
+		lsb = (const struct lvm2_super_block *)(buf + i);
+
+		if (!memcmp(lsb->magic, LVM2_MAGIC, LVM2_MAGIC_L) &&
+		    !memcmp(lsb->type,  LVM2_TYPE,  LVM2_TYPE_L)) {
+			/* This is just one of possibly many PV's */
+			*blocks = 0;
+			return 1;
+		}
+	}
+
+	return 0;
+}
+
 struct imagetype {
 	off_t		block;
 	const char	name[12];
-	int		(*identify)(const unsigned char *, unsigned long long *);
+	int		(*identify)(const void *, unsigned long long *);
 };
 
+/*
+ * Note:
+ *
+ * Minix test needs to come after ext3/ext2, since it's possible for
+ * ext3/ext2 to look like minix by pure random chance.
+ *
+ * LVM comes after all other filesystems since it's possible
+ * that an old lvm signature is left on the disk if pvremove
+ * is not used before creating the new fs.
+ */
 static struct imagetype images[] = {
 	{ 0,		"gzip",		gzip_image	},
 	{ 0,		"cramfs",	cramfs_image	},
 	{ 0,		"romfs",	romfs_image	},
 	{ 0,		"xfs",		xfs_image	},
 	{ 0,		"luks",		luks_image	},
-	{ 1,		"minix",	minix_image	},
 	{ 1,		"ext3",		ext3_image	},
 	{ 1,		"ext2",		ext2_image	},
+	{ 1,		"minix",	minix_image	},
 	{ 8,		"reiserfs",	reiserfs_image	},
 	{ 64,		"reiserfs",	reiserfs_image	},
 	{ 32,		"jfs",		jfs_image	},
+	{ 0,		"lvm2",		lvm2_image	},
+	{ 1,		"lvm2",		lvm2_image	},
 	{ -1,		"swap",		swap_image	},
 	{ 0,		"",		NULL 		}
 };
@@ -223,7 +260,7 @@ static struct imagetype images[] = {
 int identify_fs(int fd, const char **fstype,
 		unsigned long long *bytes, off_t offset)
 {
-	unsigned char buf[BLOCK_SIZE];
+	uint64_t buf[BLOCK_SIZE>>3]; /* 64-bit worst case alignment */
 	off_t cur_block = (off_t)-1;
 	struct imagetype *ip;
 	int ret;
